# Implements OpenApi runtime expressions
# See: https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#runtimeExpression

from typing import Any, List
import jsonpath_ng


class RuntimeExpressionException(Exception):
    pass


class RuntimeExpression:

    PREFIX = None
    CLASSES: List["RuntimeExpression".__class__] = []

    def to_string(self) -> str:
        return self.PREFIX

    @classmethod
    def from_string(cls: "RuntimeExpression", value: str) -> "RuntimeExpression":
        for expression_class in cls.CLASSES:
            if value.startswith(expression_class.PREFIX):
                return expression_class(value)
        raise RuntimeExpressionException(
            "Cannot parse '{}' as runtime expression".format(value))

    def __init__(self, value: str):
        if value != self.PREFIX:
            raise RuntimeExpressionException(
                "Expected {}, got {}".format(self.PREFIX, value))

    def lookup(self, url: str, method: str, status_code: int, response_body: dict) -> Any:
        raise NotImplemented()


class UrlRuntimeExpression(RuntimeExpression):

    PREFIX = "$url"

    def lookup(self, url: str, method: str, status_code: int, response_body: dict) -> Any:
        return url


RuntimeExpression.CLASSES.append(UrlRuntimeExpression)


class MethodRuntimeExpression(RuntimeExpression):

    PREFIX = "$method"

    def lookup(self, url: str, method: str, status_code: int, response_body: dict) -> Any:
        return method


RuntimeExpression.CLASSES.append(MethodRuntimeExpression)


class StatusCodeRuntimeExpression(RuntimeExpression):

    PREFIX = "$statusCode"

    def lookup(self, url: str, method: str, status_code: int, response_body: dict) -> Any:
        return status_code


RuntimeExpression.CLASSES.append(StatusCodeRuntimeExpression)


class RequestRuntimeExpression(RuntimeExpression):

    PREFIX = "$request."

    def __init__(self, value: str):
        raise NotImplemented()


RuntimeExpression.CLASSES.append(RequestRuntimeExpression)


class ResponseHeaderRuntimeExpression(RuntimeExpression):

    PREFIX = "$response.header."

    def __init__(self, value: str):
        raise NotImplemented()


RuntimeExpression.CLASSES.append(ResponseHeaderRuntimeExpression)


class ResponseQueryRuntimeExpression(RuntimeExpression):

    PREFIX = "$response.query."

    def __init__(self, value: str):
        raise NotImplemented()


RuntimeExpression.CLASSES.append(ResponseQueryRuntimeExpression)


class ResponsePathRuntimeExpression(RuntimeExpression):

    PREFIX = "$response.path."

    def __init__(self, value: str):
        raise NotImplemented()


RuntimeExpression.CLASSES.append(ResponsePathRuntimeExpression)


class ResponseBodyRuntimeExpression(RuntimeExpression):

    PREFIX = "$response.body"

    def __init__(self, value: str):
        if not value.startswith(self.PREFIX):
            raise RuntimeExpressionException(
                "Failed to parse '{}': expected prefix {}".format(value, self.PREFIX))
        json_path = value[len(self.PREFIX):]
        if json_path:
            if json_path.startswith("#"):
                json_path = json_path[1:]
            else:
                raise Exception(
                    "Failed to parse '{}': '{}' is not a local json path".format(value, json_path))
            try:
                self.json_path = jsonpath_ng.parse(json_path)
            except jsonpath_ng.exceptions.JSONPathError as e:
                raise RuntimeExpressionException(
                    "Failed to parse '{}': {}".format(json_path, e))
        else:
            self.json_path = None

    def lookup(self, url: str, method: str, status_code: int, response_body: dict) -> Any:
        if self.json_path:
            try:
                val = [i.value for i in self.json_path.find(response_body)]
            except KeyError as e:
                raise RuntimeExpressionException(
                    "Cannot find {}".format(self.json_path))
            if len(val) != 1:
                raise RuntimeExpressionException(
                    "Expected one value, got {}".format(val))
            return val[0]
        else:
            return response_body

    def to_string(self) -> str:
        if self.json_path:
            return self.PREFIX + "#" + str(self.json_path)
        else:
            return self.PREFIX


RuntimeExpression.CLASSES.append(ResponseBodyRuntimeExpression)



from dataclasses import dataclass
import json
from typing import Dict, List, Any
from .runtime_expression import RuntimeExpression

from .parse_util import assert_type, safe_dict_lookup
from enum import Enum

VERSION = '0.1'


@dataclass
class Request:
    path: str
    method: str
    headers: Dict[str, str]
    body: str

    def to_dict(self) -> dict:
        return {
            'path': self.path,
            'method': self.method,
            'headers': self.headers,
            'body': self.body,
        }

    @classmethod
    def from_dict(cls: "Request", data: Any, json_path: str) -> "Request":
        return cls(
            path=safe_dict_lookup(data, 'path', str, json_path),
            method=safe_dict_lookup(data, 'method', str, json_path),
            headers=safe_dict_lookup(data, 'headers', dict, json_path),
            body=safe_dict_lookup(data, 'body', str, json_path),
        )


class MatchType(Enum):
    JSON_SCHEMA = 'json_schema'
    EQUALITY = 'equality'
    STATUS_CODE_ONLY = 'only_code'


@dataclass
class Response:
    code: int
    content: str
    match: MatchType
    variables: Dict[str, RuntimeExpression]

    def to_dict(self):
        return {
            'code': self.code,
            'content': self.content,
            'match': self.match.value,
            'variables': {k: v.to_string() for k, v in self.variables.items()}
        }

    @classmethod
    def from_dict(cls: "Response", data: Any, json_path: str) -> "Response":
        return cls(
            code=safe_dict_lookup(data, 'code', int, json_path),
            content=safe_dict_lookup(data, 'content', str, json_path),
            match=MatchType(safe_dict_lookup(data, 'match', str, json_path)),
            variables={k: RuntimeExpression.from_string(v) for k, v in safe_dict_lookup(
                data, 'variables', dict, json_path).items()},
        )


@dataclass
class TestCase:
    request: Request
    response: Response

    def to_dict(self):
        return {
            'request': self.request.to_dict(),
            'response': self.response.to_dict(),
        }

    @classmethod
    def from_dict(cls: "TestCase", data: Any, json_path: str) -> "TestCase":
        assert_type(data, dict, json_path)
        return cls(
            request=Request.from_dict(safe_dict_lookup(
                data, 'request', dict, json_path), json_path + '.' + 'request'),
            response=Response.from_dict(safe_dict_lookup(
                data, 'response', dict, json_path), json_path + '.' + 'response'),
        )


@dataclass
class RunConfig:
    test_cases: List[TestCase]
    components: dict

    def to_dict(self):
        return {
            'version': VERSION,
            'test_cases': [i.to_dict() for i in self.test_cases],
            'components': self.components,
        }

    @classmethod
    def from_dict(cls: "RunConfig", data) -> "RunConfig":
        assert_type(data, dict, '')
        version = safe_dict_lookup(data, 'version', str, '')
        if version != VERSION:
            raise Exception(
                "Incompatible format: got {}, expected {}".format(version, VERSION))
        return cls(
            test_cases=[TestCase.from_dict(tc, 'test_cases.' + str(i))
                        for i, tc in enumerate(safe_dict_lookup(data, 'test_cases', list, ''))],
            components=safe_dict_lookup(data, 'components', dict, ''),
        )



from typing import Any, Type


class NoDefault:
    pass


def assert_type(value: Any, _type: Type, json_path: str):
    if not isinstance(value, _type):
        raise TypeError('Expected "{}", got "{}" at {}'.format(
            _type, type(value), json_path))
    return value


def safe_dict_lookup(data: dict, key: str, _type: Type, json_path: str, default=NoDefault):
    if key not in data:
        if default is not NoDefault:
            return default
        else:
            raise KeyError('Expected key "{}" at {}'.format(key, json_path))
    return assert_type(data[key], _type, json_path + '.' + key)



from unittest import TestCase
import os
import zipfile
import io
from xml.etree import ElementTree

from aas_test_engines import file

script_dir = os.path.dirname(os.path.realpath(__file__))


def in_memory_zipfile(path: str):
    buffer = io.BytesIO()
    zip = zipfile.ZipFile(buffer, "a", zipfile.ZIP_DEFLATED, False)
    for root, subdirs, files in os.walk(path):
        for file in files:
            real_path = os.path.join(root, file)
            archive_path = real_path[len(path)+1:]
            zip.write(real_path, archive_path)
    return zip


class CheckJsonTest(TestCase):

    def test_empty(self):
        result = file.check_json_data({})
        self.assertTrue(result.ok())


class CheckXmlTest(TestCase):

    def test_empty(self):
        data = ElementTree.fromstring(
            """<environment xmlns="https://admin-shell.io/aas/3/0">
            </environment>""")
        result = file.check_xml_data(data)
        self.assertTrue(result.ok())

    def test_invalid_namespace(self):
        data = ElementTree.fromstring(
            """<environment xmlns="invalid">
            </environment>""")
        result = file.check_xml_data(data)
        self.assertFalse(result.ok())
        result.dump()

class CheckAasxTest(TestCase):

    def test_empty(self):
        z = in_memory_zipfile(os.path.join(
            script_dir, 'fixtures/aasx/invalid/empty'))
        result = file.check_aasx_data(z)
        result.dump()
        self.assertFalse(result.ok())

    def test_no_rels(self):
        z = in_memory_zipfile(os.path.join(
            script_dir, 'fixtures/aasx/invalid/no_rels'))
        result = file.check_aasx_data(z)
        result.dump()
        self.assertFalse(result.ok())

    def test_invalid_rels(self):
        z = in_memory_zipfile(os.path.join(
            script_dir, 'fixtures/aasx/invalid/invalid_rels'))
        result = file.check_aasx_data(z)
        result.dump()
        self.assertFalse(result.ok())

    def test_unknown_filetype(self):
        z = in_memory_zipfile(os.path.join(
            script_dir, 'fixtures/aasx/invalid/unknown_filetype'))
        result = file.check_aasx_data(z)
        result.dump()
        self.assertFalse(result.ok())

    def test_minimal(self):
        z = in_memory_zipfile(os.path.join(
            script_dir, 'fixtures/aasx/valid/simple'))
        result = file.check_aasx_data(z)
        result.dump()
        self.assertTrue(result.ok())


class SupportedVersionTest(TestCase):

    def test_invoke(self):
        s = file.supported_versions()
        for i in s:
            print(i)
        self.assertIn(file.latest_version(), s)

class AasCoreTestCase(TestCase):

    def is_blacklisted(self, path):
        blacklist = [
            'UnexpectedAdditionalProperty',
            'Double/lowest.',
            'Double/max.',
            'Float/largest_normal.',
            'ConstraintViolation/Reference/',
            'ConstraintViolation/reference',
            'ConstraintViolation/SubmodelElementList',
            'ConstraintViolation/submodelElementList',
            'basicEventElement/messageTopic.xml',
            'TypeViolation/langStringNameType/text.xml',
            'TypeViolation/blob/value.xml',
            'TypeViolation/langStringShortNameTypeIec61360/text.xml',
            'TypeViolation/assetInformation/assetType.xml',
            'TypeViolation/assetInformation/globalAssetId.xml',
            'TypeViolation/submodelElementList/orderRelevant.xml',
            'TypeViolation/conceptDescription/administration.xml',
            'TypeViolation/conceptDescription/id.xml',
            'TypeViolation/administrativeInformation/templateId.xml',
            'TypeViolation/specificAssetId/value.xml',
            'TypeViolation/specificAssetId/name.xml',
            'TypeViolation/langStringPreferredNameTypeIec61360/text.xml',
            'TypeViolation/langStringTextType/text.xml',
            'TypeViolation/submodel/administration.xml',
            'TypeViolation/submodel/id.xml',
            'TypeViolation/dataSpecificationIec61360/symbol.xml',
            'TypeViolation/dataSpecificationIec61360/value.xml',
            'TypeViolation/dataSpecificationIec61360/sourceOfDefinition.xml',
            'TypeViolation/dataSpecificationIec61360/valueFormat.xml',
            'TypeViolation/dataSpecificationIec61360/unit.xml',
            'TypeViolation/langStringDefinitionTypeIec61360/text.xml',
            'TypeViolation/assetAdministrationShell/administration.xml',
            'TypeViolation/assetAdministrationShell/id.xml',
            'TypeViolation/key/value.xml',
            'TypeViolation/valueReferencePair/value.xml',
            'TypeViolation/levelType/min.xml',
            'TypeViolation/levelType/max.xml',
            'TypeViolation/levelType/typ.xml',
            'TypeViolation/levelType/nom.xml',
        ]
        for i in blacklist:
            if i in path:
                return True
        return False

    def _run(self, dirname: str, check):
        print(f"Testing {dirname}, this might take a few minutes...")
        valid_accepted = 0
        valid_rejected = 0
        invalid_accepted = 0
        invalid_rejected = 0
        skipped = 0
        for root, dirs, files in os.walk(os.path.join(script_dir, f'fixtures/aas-core3.0-testgen/test_data/{dirname}/ContainedInEnvironment'), topdown=False):
            for name in files:
                path_in = os.path.join(root, name)
                if self.is_blacklisted(path_in):
                    skipped += 1
                    continue
                with open(path_in) as f:
                    errors = check(f)
                if errors.ok():
                    if 'Expected' in path_in:
                        valid_accepted += 1
                    else:
                        invalid_accepted += 1
                else:
                    if 'Expected' in path_in:
                        valid_rejected += 1
                    else:
                        invalid_rejected += 1
        print("valid, accepted",     valid_accepted)
        print("invalid, rejected",   invalid_rejected)
        print("valid, rejected",     valid_rejected)
        print("invalid, accepted",   invalid_accepted)
        print("skipped",             skipped)

    def test_json(self):
        self._run('Json', file.check_json_file)

    def test_xml(self):
        self._run('Xml', file.check_xml_file)



from typing import List

class Resolver:

    def __init__(self, root_document: any):
        self.root_document = root_document

    def lookup(self, pointer: str):
        if pointer.startswith('#/'):
            pointer = pointer[2:]
        components = pointer.split("/")
        return self._lookup(components, self.root_document)

    def _lookup(self, components: List[str], data: any):
        if not components:
            return data
        key = components.pop(0)
        return self._lookup(components, data[key])


import pydot
import openapi


def to_graph(api: openapi.OpenApi):

    graph = pydot.Dot("link_dependencies", graph_type="digraph")

    for path in api.paths:
        for operation in path.operations:
            graph.add_node(pydot.Node(
                operation.operation_id,
                label=operation.operation_id + "\n" + operation.method.upper() + " " +
                path.path,
                shape="rect",
            ))
            for param in operation.parameters:
                if param.source_link:
                    graph.add_edge(pydot.Edge(
                        param.source_link.source_operation.operation_id,
                        operation.operation_id,
                        color="blue",
                        label=param.name,
                    ))

    return graph



from unittest import TestCase

from aas_test_engines import api


class ApiTestCase(TestCase):

    def test_simple(self):
        pass


class SupportedVersionsTest(TestCase):

    def test_list(self):
        s = api.supported_versions()
        for i in s:
            print(i)
        self.assertIn(api.latest_version(), s)



from typing import Dict, List, Optional, Set, Tuple

from dataclasses import dataclass, field
import hashlib
import json
from urllib.parse import urlencode
import os
import base64

from . import runconf
from . import openapi


@dataclass
class RequestParams:
    path_params: Dict[str, str] = field(default_factory=dict)
    body: str = ''
    headers: Dict[str, str] = field(default_factory=dict)
    query_params: Dict[str, str] = field(default_factory=dict)

    class Discard:
        pass

    def to_request(self, path: str, method: str) -> runconf.Request:
        for key, value in self.path_params.items():
            path = path.replace('{' + key + '}', value)
        if self.query_params:
            path += "?" + urlencode(self.query_params)
        return runconf.Request(
            path=path,
            method=method,
            # TODO: content type should not be hardcoded
            headers={**self.headers, 'content-type': 'application/json'},
            body=self.body,
        )


class ParameterOptions:
    def __init__(self, values: List[str]) -> None:
        self.values = values

    def inject(self, request: RequestParams, value: str):
        raise NotImplemented()


class QueryParameterOption(ParameterOptions):
    def __init__(self, key: str, values: str) -> None:
        super().__init__(values)
        self.key = key

    def inject(self, request: RequestParams, value: str):
        request.query_params[self.key] = value


class PathParameterOption(ParameterOptions):
    def __init__(self, key: str, values: str) -> None:
        super().__init__(values)
        self.key = key

    def inject(self, request: RequestParams, value: str):
        request.path_params[self.key] = value


class JsonBodyParameterOption(ParameterOptions):
    def __init__(self, values: str) -> None:
        super().__init__(values)

    def inject(self, request: RequestParams, value: str):
        request.body = json.dumps(value)


def generate_link_id(source_op: openapi.Operation, param: str, target_op: openapi.Operation) -> str:
    id_ = source_op.operation_id + "_" + param + "_" + target_op.operation_id
    return hashlib.sha256(id_.encode()).hexdigest()


def generate_valid_samples(schema: Optional[dict]):
    if 'enum' in schema:
        return schema['enum']
    if '$ref' in schema:
        ref = schema['$ref']
        samples_dir = os.path.dirname(
            os.path.realpath(__file__)) + '/../data/api/samples/'
        # TODO: make this configurable
        path = {
            '#/components/schemas/AssetAdministrationShell': 'aas.json',
            '#/components/schemas/AssetInformation': 'asset_info.json',
            '#/components/schemas/Reference': 'model_reference.json',
            '#/components/schemas/Submodel': 'submodel.json',
            '#/components/schemas/SubmodelElement': 'submodel_element.json',
            '#/components/schemas/OperationRequest': 'operation_request.json',
            '#/components/schemas/AssetAdministrationShellDescriptor': 'shell_descriptor.json',
            '#/components/schemas/SubmodelDescriptor': 'submodel_descriptor.json',
            '#/components/schemas/ConceptDescription': 'empty.json',
            '#/components/schemas/Identifier': 'empty.json',
            '#/components/schemas/IdentifierKeyValuePair': 'empty.json',
            '#/components/schemas/SubmodelMetadata': 'empty.json',
            '#/components/schemas/SubmodelValue': 'empty.json',
            '#/components/schemas/SubmodelElementMetadata': 'empty.json',
            '#/components/schemas/SubmodelElementValue': 'empty.json',
            '#/components/schemas/OperationRequestValueOnly': 'empty.json',
        }.get(ref)
        if path:
            # TODO: check if example matches the schema
            return [json.load(open(samples_dir + path, "rb"))]
        else:
            raise Exception("Unknown ref " + schema['$ref'])
    type_ = schema.get('type', 'object')
    if type_ == 'array':
        return [generate_valid_samples(schema['items'])]
    elif type_ == 'boolean':
        return [True, False]
    elif type_ == 'string':
        return ["random_string"]
    else:
        # Note we could pick random values for types 'string' and 'integer', too.
        # However, these are in most cases ids, which should be captured from other requests.
        # One should use OpenApi links for this purpose.
        raise Exception("Unknown schema type {}".format(type_))


def generate_invalid_samples(schema: dict):
    # TODO: this could be more elaborated ;)
    return ["INVALID"]


def extract_variables(operation: openapi.Operation, response: openapi.Response):
    # TODO: it is not needed to extract ALL variables, but only the ones we actually need for other tests, later
    # Furthermore, a lot of variables will have the same value, so they should not be evaluated multiple times
    vars = {}
    for link in response.links:
        for param, expression in link.parameters.items():
            id_ = generate_link_id(operation, param, link.target_operation)
            vars[id_] = expression
    return vars


def generate_positive_tests(test_cases: List[runconf.TestCase], path: str, operation: openapi.Operation, response: openapi.Response) -> List[RequestParams]:

    options: List[ParameterOptions] = []

    for i in operation.parameters:
        if not i.required:
            # TODO: should be generated, too
            options.append(PathParameterOption(
                key=i.name,
                values=[RequestParams.Discard]
            ))
            continue
        if i.source_link:
            id_ = generate_link_id(i.source_link.source_operation, i.name, operation)
            base64url_encoded = i.schema.get('format') == 'byte'
            if base64url_encoded:
                values = ['!{' + id_ + '_base64}']
            else:
                values = ['!{' + id_ + '}']
            # TODO: if base64_urlencoded, add negative test which does not encode
        else:
            # TODO: escape strings of the form !{*}, to avoid clash with variables above
            values = generate_valid_samples(i.schema)
        if i.position == openapi.ParameterPosition.PATH:
            options.append(PathParameterOption(
                key=i.name,
                values=values,
            ))
        elif i.position == openapi.ParameterPosition.QUERY:
            options.append(QueryParameterOption(
                key=i.name,
                values=values,
            ))
        else:
            raise Exception(
                "Parameter position {} not implemented".format(i.position))

    if operation.request_body and operation.request_body.schema:
        options.append(JsonBodyParameterOption(
            generate_valid_samples(operation.request_body.schema)
        ))

    variables = extract_variables(operation, response)

    all_params: List[RequestParams] = []

    if options:
        num_test_cases = max([len(i.values) for i in options])
        for i in range(num_test_cases):
            params = RequestParams()
            for option in options:
                value = option.values[i % len(option.values)]
                if value != RequestParams.Discard:
                    option.inject(params, value)

            req = params.to_request(path, operation.method)
            if response.schema is not None:
                res = runconf.Response(
                    code=response.code,
                    content=json.dumps(response.schema),
                    match=runconf.MatchType.JSON_SCHEMA,
                    variables=variables
                )
            else:
                res = runconf.Response(
                    code=response.code,
                    content='',
                    match=runconf.MatchType.STATUS_CODE_ONLY,
                    variables=variables
                )
            test_cases.append(runconf.TestCase(req, res))
            all_params.append(params)
    else:
        params = RequestParams()
        req = params.to_request(path, operation.method)
        res = runconf.Response(
            code=response.code,
            content=json.dumps(response.schema),
            match=runconf.MatchType.JSON_SCHEMA,
            variables=variables,
        )
        test_cases.append(runconf.TestCase(req, res))
        all_params.append(params)

    # We need these for negative test case generation
    return all_params


def generate_negative_tests(test_cases: List[runconf.TestCase],  positive_params: List[RequestParams], path: str, operation: openapi.Operation, response: openapi.Response) -> runconf.TestCase:

    if not positive_params:
        raise Exception("Need positive params as basis for negative tests")

    options: List[ParameterOptions] = []
    for i in operation.parameters:
        values = generate_invalid_samples(i.schema)
        if i.required:
            values += [RequestParams.Discard]
        if i.position == openapi.ParameterPosition.PATH:
            options.append(PathParameterOption(
                key=i.name,
                values=values,
            ))
        elif i.position == openapi.ParameterPosition.QUERY:
            options.append(QueryParameterOption(
                key=i.name,
                values=values,
            ))
        else:
            raise Exception(
                "Parameter position {} not implemented".format(i.position))

    if operation.request_body:
        options.append(JsonBodyParameterOption(
            generate_invalid_samples(operation.request_body.schema)
        ))

    positive_params_idx = 0
    for option in options:
        for value in option.values:
            params = positive_params[positive_params_idx]
            if value != RequestParams.Discard:
                option.inject(params, value)
            positive_params_idx = (
                positive_params_idx + 1) % len(positive_params)
            req = params.to_request(path, operation.method)
            res = runconf.Response(
                code=response.code,
                content=json.dumps(response.schema),
                match=runconf.MatchType.JSON_SCHEMA,
                variables={},
            )
            test_cases.append(runconf.TestCase(req, res))


def generate_tests(test_cases: List[runconf.TestCase], path: str, operation: openapi.Operation) -> runconf.TestCase:
    # print("Generating tests for {} {}".format(operation.method, path))
    positive_response = None
    negative_response = None
    for response in operation.responses:
        if response.code >= 200 and response.code < 300:
            if positive_response:
                # print("Ignoring multiple positives for " + path)
                continue
            positive_response = response
        elif response.code >= 400 and response.code < 500:
            if negative_response:
                # print("Ignoring multiple negatives for " + path)
                continue
            negative_response = response
    if not positive_response:
        raise Exception("Must have a positive response")
    positive_params = generate_positive_tests(
        test_cases, path, operation, positive_response)
    if not negative_response:
        # print("No negative response given for {} {}, guessing one".format(operation.method, path))
        negative_response = openapi.Response(
            code=400,
            schema={},
            links=[]
        )
    generate_negative_tests(test_cases, positive_params,
                            path, operation, negative_response)


def has_all_params(operation: openapi.Operation, already_tested: Set[str]) -> bool:
    for param in operation.parameters:
        if param.source_link:
            if param.source_link.source_operation.operation_id not in already_tested:
                return False
    return True


def get_next(api: openapi.OpenApi, already_tested: Set[str], tag_filter: Set[str]) -> Tuple[openapi.Path, openapi.Operation]:

    def find(delete_only: bool):
        it = reversed(api.paths) if delete_only else api.paths
        for path in it:
            for operation in path.operations:
                if (operation.method == 'delete') != delete_only:
                    continue
                if not (operation.tags & tag_filter):
                    continue
                if operation.operation_id in already_tested:
                    continue
                if has_all_params(operation, already_tested):
                    return path, operation

    result = find(False)
    if result:
        return result
    result = find(True)
    if result:
        return result
    raise Exception("No more operations whose parameters are available. Maybe you have a cyclic dependency within your links?")


def generate(api: openapi.OpenApi, tag_filter=Set[str]):
    test_cases: List[runconf.TestCase] = []
    num_ops = 0
    for path in api.paths:
        for op in path.operations:
            if op.tags & tag_filter:
                num_ops += 1
    already_tested: Set[str] = set()
    while len(already_tested) != num_ops:
        next_path, next_operation = get_next(api, already_tested, tag_filter)
        already_tested.add(next_operation.operation_id)
        generate_tests(test_cases, next_path.path, next_operation)
    return runconf.RunConfig(
        components=api.components,
        test_cases=test_cases,
    )



from typing import Dict, Set, Generator
from .exception import AasTestToolsException
from .result import AasTestResult

from ._api import openapi
from ._api import generate
from ._api import runconf
from ._api import run

import os
from yaml import safe_load


class AasSpec:

    def __init__(self, api: openapi.OpenApi, tags: Set[str]) -> None:
        self.api = api
        self.tags = tags


def _find_specs() -> Dict[str, AasSpec]:
    result = {}
    script_dir = os.path.dirname(os.path.realpath(__file__))
    data_dir = os.path.join(script_dir, 'data', 'api')
    for i in os.listdir(data_dir):
        path = os.path.join(data_dir, i)
        if not i.endswith('.yml'):
            continue
        spec = safe_load(open(path, "rb"))
        api = openapi.OpenApi.from_dict(spec)
        suites = set()
        for path in api.paths:
            for operation in path.operations:
                suites.update(operation.tags)
        result[i[:-4]] = AasSpec(api, suites)
    return result


_specs = _find_specs()

_DEFAULT_VERSION = '3.0'


def _get_spec(version: str) -> AasSpec:
    try:
        return _specs[version]
    except KeyError:
        raise AasTestToolsException(
            f"Unknown version {version}, must be one of {supported_versions()}")


def generate_tests(version: str = _DEFAULT_VERSION, suites: Set[str] = None) -> runconf.RunConfig:
    spec = _get_spec(version)
    if suites is None:
        suites = spec.tags
    if not spec.tags.issuperset(suites):
        raise AasTestToolsException(f"Unknown suites {suites}, must be in {spec.tags}")
    conf = generate.generate(spec.api, suites)
    return conf


def execute_tests(conf: runconf.RunConfig, server: str, dry: bool = False) -> Generator[AasTestResult, None, None]:
    yield from run.run(conf, server, dry)


def supported_versions():
    return {ver: spec.tags for ver, spec in _specs.items()}


def latest_version():
    return _DEFAULT_VERSION



from typing import List, Dict, TextIO, Union, Any, Set, Optional
import os
import json
from yaml import safe_load

from .exception import AasTestToolsException
from .result import AasTestResult, Level
from .data_types import validators

from xml.etree import ElementTree
from json_schema_tool.schema import SchemaValidator, SchemaValidator, ValidationConfig, ParseConfig, SchemaValidationResult, KeywordValidationResult, parse_schema
from json_schema_tool.types import JsonType
from json_schema_tool.exception import PreprocessorException
from zipfile import ZipFile

JSON = Union[str, int, float, bool, None, Dict[str, Any], List[Any]]


class AasSchema:

    def __init__(self, validator: SchemaValidator):
        self.validator = validator


def _find_schemas() -> Dict[str, any]:
    result = {}
    script_dir = os.path.dirname(os.path.realpath(__file__))
    data_dir = os.path.join(script_dir, 'data', 'file')
    for i in os.listdir(data_dir):
        path = os.path.join(data_dir, i)
        if not i.endswith('.yml'):
            continue
        schema = safe_load(open(path, "rb"))
        config = ParseConfig(
            format_validators=validators
        )
        validator = parse_schema(schema, config)
        result[i[:-4]] = AasSchema(validator)
    return result


_schemas = _find_schemas()
_DEFAULT_VERSION = '3.0'

def supported_versions():
    return list(_schemas.keys())

def latest_version():
    return _DEFAULT_VERSION

def _get_schema(version: str) -> AasSchema:
    try:
        return _schemas[version]
    except KeyError:
        raise AasTestToolsException(
            f"Unknown version {version}, must be one of {supported_versions()}")

def _map_error(parent: AasTestResult, error: SchemaValidationResult):
    for i in error.keyword_results:
        if i.ok(): continue
        kw_result = AasTestResult(i.error_message, '', Level.ERROR)
        for j in i.sub_schema_results:
            _map_error(kw_result, j)
        parent.append(kw_result)

def check_json_data(data: any, version: str = _DEFAULT_VERSION) -> AasTestResult:
    schema = _get_schema(version)
    result = AasTestResult('Check JSON', '', Level.INFO)
    error = schema.validator.validate(data)
    _map_error(result, error)
    return result


def check_json_file(file: TextIO, version: str = _DEFAULT_VERSION) -> AasTestResult:
    data = json.load(file)
    return check_json_data(data, version)


def _get_model_type(el: ElementTree.Element, expected_namespace: str):
    model_type = el.tag[len(expected_namespace):]
    model_type = model_type[0].upper() + model_type[1:]
    return model_type

def _get_single_child(el: ElementTree.Element) -> ElementTree.Element:
    if len(el) != 1:
        raise Exception("DataSpecificationContent must have exactly one child")
    return el[0]



def check_xml_data(data: ElementTree, version: str = _DEFAULT_VERSION) -> AasTestResult:
    expected_namespace = '{https://admin-shell.io/aas/3/0}'

    def preprocess(data: ElementTree.Element, validator: SchemaValidator) -> JSON:

        if isinstance(data, (dict, list, str, bool, int, float)) or data is None:
            return data

        if not data.tag.startswith(expected_namespace):
            raise PreprocessorException(f"invalid namespace, got '{data.tag}'")

        types = validator.get_types()

        if types == {JsonType.OBJECT}:

            # Special handling for data specification content
            if data.tag.endswith('dataSpecificationContent'):
                data = _get_single_child(data)

            result = {}
            result['modelType'] = _get_model_type(data, expected_namespace)
            for child in data:
                if not child.tag.startswith(expected_namespace):
                    raise PreprocessorException(f"invalid namespace, got {child.tag}")
                tag = child.tag[len(expected_namespace):]
                if result['modelType'] == 'OperationVariable' and tag == 'value':
                    result[tag] = _get_single_child(child)
                else:
                    result[tag] = child
            return result
        elif types == {JsonType.ARRAY}:
            result = []
            for child in data:
                result.append(child)
            return result
        elif types == {JsonType.STRING}:
            return data.text or ""
        elif types == {JsonType.BOOLEAN}:
            return data.text == 'true'
        else:
            raise Exception(f"Unknown type {types} at {validator.pointer}")
    schema = _get_schema(version)
    config = ValidationConfig(preprocessor=preprocess)
    error = schema.validator.validate(data, config)
    result = AasTestResult('Check XML', '', Level.INFO)
    _map_error(result, error)
    return result


def check_xml_file(file: TextIO, version: str = _DEFAULT_VERSION) -> AasTestResult:
    data = ElementTree.fromstring(file.read())
    return check_xml_data(data, version)


NS_CONTENT_TYPES = "{http://schemas.openxmlformats.org/package/2006/content-types}"
NS_RELATIONSHIPS = "{http://schemas.openxmlformats.org/package/2006/relationships}"

TYPE_AASX_ORIGIN = 'http://www.admin-shell.io/aasx/relationships/aasx-origin'
TYPE_AASX_SPEC = 'http://www.admin-shell.io/aasx/relationships/aas-spec'
TYPE_AASX_SUPPL = 'http://www.admin-shell.io/aasx/relationships/aas-suppl'
TYPE_THUMBNAIL = 'http://schemas.openxmlformats.org/package/2006/relationships/metadata/thumbnail'


class Relationship:

    def __init__(self, type: str, target: str) -> None:
        self.type = type
        self.target = target
        self.sub_rels: List[Relationship] = []

    def sub_rels_by_type(self, type: str) -> List["Relationship"]:
        return [i for i in self.sub_rels if i.type == type]

    def dump(self, indent=0):
        type_suffix = self.type.split("/")[-1]
        print("  " * indent + f"{self.target} [{type_suffix}]")
        for sub_rel in self.sub_rels:
            sub_rel.dump(indent + 1)


def _check_content_type(zipfile: ZipFile) -> AasTestResult:
    content_types_xml = '[Content_Types].xml'
    result = AasTestResult(f'Checking {content_types_xml}', content_types_xml)
    try:
        with zipfile.open(content_types_xml, 'r') as f:
            content_types = ElementTree.parse(f)
            expected_tag = f'{NS_CONTENT_TYPES}Types'
            if content_types.getroot().tag != expected_tag:
                result.append(
                    AasTestResult(f"root must have tag {expected_tag}, got {content_types.getroot().tag}", content_types_xml, Level.ERROR))
    except KeyError:
        result.append(AasTestResult(
            f"{content_types_xml} not found", content_types_xml, Level.ERROR))

    return result


def _scan_relationships(zipfile: ZipFile, parent_rel: Relationship, rels_file: TextIO, visited_targets: Set[str]) -> AasTestResult:
    relationships = ElementTree.parse(rels_file).getroot()
    expected_tag = f"{NS_RELATIONSHIPS}Relationships"
    if relationships.tag != expected_tag:
        return AasTestResult(f'Invalid root tag {relationships.tag}, expected {expected_tag}', relationships.tag, Level.ERROR)

    result = AasTestResult(f"Checking relationships", relationships.tag)
    for idx, rel in enumerate(relationships):
        if rel.tag != f"{NS_RELATIONSHIPS}Relationship":
            result.append(AasTestResult(
                f'Invalid tag {rel.tag}', str(idx), Level.ERROR))
            continue
        try:
            type = rel.attrib['Type']
            target = rel.attrib['Target']
        except KeyError as e:
            result.append(AasTestResult(
                f'Attribute {e} is missing', str(idx), Level.ERROR))
            continue

        if target.startswith('/'):
            target = target[1:]

        sub_dir, file = os.path.split(target)
        sub_rel = Relationship(type, target)
        parent_rel.sub_rels.append(sub_rel)
        if target in visited_targets:
            result.append(AasTestResult(
                'Recursive relationship', str(idx), Level.ERROR))
            continue
        visited_targets.add(target)
        if target not in zipfile.namelist():
            result.append(AasTestResult(
                f'Relationship has non-existing target {target}', str(idx), Level.ERROR))
            continue
        try:
            with zipfile.open(f"{sub_dir}/_rels/{file}.rels", "r") as f:
                r = _scan_relationships(zipfile, sub_rel, f, visited_targets)
                result.append(r)
        except KeyError:
            # No further sub-relationships
            pass
        result.append(AasTestResult(
            f'Relationship {target} is ok', str(idx), Level.INFO))

    return result


def _check_relationships(zipfile: ZipFile, root_rel: Relationship) -> AasTestResult:
    result = AasTestResult('Checking relationships', '')
    visited_targets = set()
    try:
        ROOT_REL_PATH = "_rels/.rels"
        with zipfile.open(ROOT_REL_PATH, "r") as f:
            r = _scan_relationships(zipfile, root_rel, f, visited_targets)
            result.append(r)
    except KeyError:
        result.append(AasTestResult(
            f"{ROOT_REL_PATH} does not exist", ROOT_REL_PATH, Level.ERROR))
    return result


def _check_files(zipfile: ZipFile, root_rel: Relationship, version: str) -> AasTestResult:
    result = AasTestResult('Checking files', '')
    for aasx_origin in root_rel.sub_rels_by_type(TYPE_AASX_ORIGIN):
        sub_result = AasTestResult(f'Checking {aasx_origin.target}', aasx_origin.target)
        for aasx_spec in aasx_origin.sub_rels_by_type(TYPE_AASX_SPEC):
            try:
                with zipfile.open(aasx_spec.target) as f:
                    if aasx_spec.target.endswith('.xml'):
                        r = check_xml_file(f, version)
                    elif aasx_spec.target.endswith('.json'):
                        r = check_json_file(f, version)
                    else:
                        r = AasTestResult('Unknown filetype', aasx_spec.target, Level.WARNING)
                    sub_result.append(r)
            except KeyError:
                return AasTestResult("File does not exist")
        result.append(sub_result)
    return result


def check_aasx_data(zipfile: ZipFile, version: str = _DEFAULT_VERSION) -> AasTestResult:

    result = AasTestResult('Checking AASX package', '')

    result.append(_check_content_type(zipfile))
    if not result.ok():
        return result

    root_rel = Relationship('ROOT', '/')
    result.append(_check_relationships(zipfile, root_rel))
    if not result.ok():
        return result

    result.append(_check_files(zipfile, root_rel, version))
    if not result.ok():
        return result

    return result


def check_aasx_file(file: TextIO, version: str = _DEFAULT_VERSION) -> AasTestResult:
    zipfile = ZipFile(file)
    return check_aasx_data(zipfile, version)



#! /usr/bin/env python3

from dataclasses import dataclass
from .runconf import RunConfig, TestCase, Response, MatchType
import json
from typing import Dict, Optional, Generator
from string import Template
from aas_test_engines.result import AasTestResult, Level

from .runtime_expression import RuntimeExpressionException

import requests
import jsonschema
import base64

from aas_test_engines.exception import AasTestToolsException


def _check_server(server: str) -> AasTestResult:
    result = AasTestResult(f'Check {server}')
    try:
        requests.get(server)
        result.append(AasTestResult('OK', '', Level.INFO))
    except requests.exceptions.RequestException as e:
        result.append(AasTestResult('Failed to reach: {}'.format(e), '', Level.ERROR))
    return result


def _check_response(test_case: TestCase, actual: requests.models.Response, data: Optional[dict], config: RunConfig, result: AasTestResult):
    expected = test_case.response
    if actual.status_code == expected.code:
        result.append(AasTestResult(f'Got status code {expected.code}'))
    else:
        result.append(AasTestResult(f"invalid status code: expected {expected.code}, got {actual.status_code}", '', Level.ERROR))

    if expected.match == MatchType.JSON_SCHEMA:
        schema = json.loads(test_case.response.content)
        schema['components'] = config.components
        try:
            jsonschema.validate(instance=data, schema=schema)
        except jsonschema.exceptions.ValidationError as e:
            result.append(AasTestResult(f"Invalid response: {e.args[0]}", '', Level.ERROR))
    elif expected.match == MatchType.STATUS_CODE_ONLY:
        pass
    else:
        raise AasTestToolsException(f"Unknown match type {expected.match}")


class TemplateWithNumericIds(Template):
    idpattern = r'(?a:[a-z0-9]+?(_base64))'
    delimiter = '!'


def inject_variables(value: str, variables: Dict[str, str]) -> str:
    return TemplateWithNumericIds(value).substitute(variables)


def _run_test_case(test_case: TestCase, server: str, dry: bool, variables: Dict[str, str], config: RunConfig) -> AasTestResult:
    url = server + test_case.request.path
    method = test_case.request.method

    try:
        url = inject_variables(url, variables)
        data = inject_variables(test_case.request.body, variables)
    except KeyError as e:
        m = "Failed to substitute variable {}, considering test case as failed".format(e.args[0])
        return AasTestResult(m, '', Level.ERROR)

    result = AasTestResult("{} {}".format(test_case.request.method.upper(), url))

    if dry:
        for name in test_case.response.variables.keys():
            variables[name] = 'dummy_value'
        result.append(AasTestResult('Skipped', '', Level.WARNING))
        return result
    else:
        response = requests.request(
            method=method,
            url=url,
            data=data,
            headers=test_case.request.headers
        )
        try:
            data = response.json()
        except:
            data = None

        _check_response(test_case, response, data, config, result)
        for name, expression in test_case.response.variables.items():
            try:
                # TODO: str() will fail for non primitive types
                value = str(expression.lookup(url, method, response.status_code, data))
                variables[name] = value
                # TODO: we actually do not need ALL variables as encoded b64, later
                variables[name+'_base64'] = base64.urlsafe_b64encode(value.encode()).decode().replace('=', '')
            except RuntimeExpressionException:
                m = f"Failed to fetch variable {name}, subsequent test cases might fail"
                result.append(AasTestResult(m, '', Level.WARNING))

        # print("--- Additional info:")
        # print(response.content)
        # print("---")
    return result

def run(config: RunConfig, server: str, dry: bool = False) -> Generator[AasTestResult, None, None]:
    if not dry:
        result = _check_server(server)
        yield result
        if not result.ok():
            return
    variables: Dict[str, str] = {}
    for test_case in config.test_cases:
        yield _run_test_case(test_case, server, dry, variables, config)



import argparse
import sys
from aas_test_engines import api, file


def run_file_test(argv):
    parser = argparse.ArgumentParser(description='Runs test cases')
    parser.add_argument('file',
                        type=argparse.FileType('r'),
                        help='the file to check')
    args = parser.parse_args(argv)
    result = file.check_aasx_file(args.file)
    print(result)


def run_api_test(argv):
    parser = argparse.ArgumentParser(description='Runs test cases')
    parser.add_argument('server',
                        type=str,
                        help='server to run the tests against')
    parser.add_argument('--dry',
                        action='store_true',
                        help="dry run, do not send requests")
    parser.add_argument('--suite',
                        type=str,
                        help='selected test suite')
    args = parser.parse_args(argv)
    if args.suite:
        suites = set([args.suite])
    else:
        suites = None
    tests = api.generate_tests(suites=suites)
    for result in api.execute_tests(tests, args.server, args.dry):
        result.dump()


commands = {
    'file': run_file_test,
    'api': run_api_test,
}

if len(sys.argv) <= 1:
    print(f"Usage: {sys.argv[0]} COMMAND OPTIONS...")
    print("Available commands:")
    print("  file     Check a file for compliance.")
    print("  api      Check a server instance for compliance.")
    exit(1)

command = sys.argv[1]

if command not in commands:
    print(f"Unknown command '{command}', must be one of {', '.join(commands)}")
    exit(1)

remaining_args = sys.argv[2:]
commands[command](remaining_args)



from typing import Any, Optional, List, Dict, Set
from dataclasses import dataclass
from enum import Enum
from .parse_util import assert_type, safe_dict_lookup
from .runtime_expression import RuntimeExpression

from .resolver import Resolver
import warnings

@dataclass
class Info:
    title: str

    @classmethod
    def from_dict(cls, data: Any, json_path: str) -> "Info":
        return cls(
            title=safe_dict_lookup(data, 'title', str, json_path)
        )


class ParameterPosition(Enum):
    QUERY = "query"
    HEADER = "header"
    PATH = "path"
    COOKIE = "cookie"


class ParameterStyle(Enum):
    FORM = "form"
    SIMPLE = "simple"


@dataclass
class Parameter:
    name: str
    position: ParameterPosition
    required: bool
    style: ParameterStyle
    explode: bool
    schema: dict
    source_link: Optional["BackwardLink"] = None

    @classmethod
    def from_dict(cls: "Parameter", data: Any, json_path: str, resolver: Resolver) -> "Parameter":
        if '$ref' in data:
            ref = safe_dict_lookup(data, '$ref', str, json_path)
            data = resolver.lookup(ref)
        return cls(
            name=safe_dict_lookup(data, 'name', str, json_path),
            position=ParameterPosition(
                safe_dict_lookup(data, 'in', str, json_path)),
            required=safe_dict_lookup(data, "required", bool, json_path, True),
            style=ParameterStyle(safe_dict_lookup(
                data, 'style', str, json_path, ParameterStyle.SIMPLE.value)),
            explode=safe_dict_lookup(data, 'explode', bool, json_path, False),
            schema=safe_dict_lookup(data, 'schema', dict, json_path),
        )


@dataclass
class RequestBody:
    description: str
    schema: Optional[Dict]
    required: bool

    @classmethod
    def from_dict(cls: "RequestBody", data: Any, json_path: str) -> "RequestBody":
        assert_type(data, dict, json_path)
        content = safe_dict_lookup(data, 'content', dict, json_path)
        json_content_type = 'application/json'
        if len(content.keys()) != 1:
            warnings.warn(f"Ignoring some content types at {json_path}")
        if json_content_type not in content.keys():
            warnings.warn(f"Support for non-json request bodies not implemented, at {json_path}")
            json_content = {}
        else:
            json_content = safe_dict_lookup(content, json_content_type, dict, json_path)
        return cls(
            description=safe_dict_lookup(data, 'description', str, json_path),
            required=safe_dict_lookup(data, 'required', bool, json_path, True),
            schema=safe_dict_lookup(
                json_content, 'schema', dict, json_path, None)
        )


@dataclass
class Link:
    name: str
    target_operation: "Operation"
    parameters: Dict[str, RuntimeExpression]

    @classmethod
    def from_dict(cls: "Link", name: str, data: Any, json_path: str):
        assert_type(data, dict, json_path)
        return cls(
            name=name,
            target_operation=safe_dict_lookup(
                data, 'operationId', str, json_path),
            parameters={k: RuntimeExpression.from_string(v) for k, v in safe_dict_lookup(
                data, "parameters", dict, json_path).items()},
        )


@dataclass
class BackwardLink:
    source_operation: "Operation"
    source_response: "Response"
    param_query: str


@dataclass
class Response:
    code: int
    schema: Optional[Dict]
    links: List[Link]

    @classmethod
    def from_dict(cls: "Response", code: str, data: Any, json_path) -> "Response":
        content = safe_dict_lookup(data, 'content', dict, json_path, None)
        schema = None
        if content is not None:
            json_content_type = 'application/json'
            if len(content.keys()) != 1:
                warnings.warn(f"Ignoring some content types at {json_path}")
            if json_content_type not in content.keys():
                warnings.warn(f"Support for non-json request bodies not implemented, at {json_path}")
                json_content = {}
            else:
                json_content = safe_dict_lookup(content, json_content_type, dict, json_path)
            schema = safe_dict_lookup(
                json_content, 'schema', dict, json_path, None)

        links = safe_dict_lookup(data, 'links', dict, json_path, None)
        if links:
            links = [Link.from_dict(k, v, json_path + '.' + k)
                     for k, v in links.items()]
        else:
            links = []

        return cls(
            code= 400 if code == 'default' else int(code),
            schema=schema,
            links=links,
        )


@dataclass
class Operation:
    operation_id: str
    summary: str
    method: str
    parameters: List[Parameter]
    request_body: Optional[RequestBody]
    responses: List[Response]
    tags: Set[str]

    @classmethod
    def from_dict(cls: "Operation", method: str, data: Any, json_path: str, resolver: Resolver) -> "Operation":
        assert_type(data, dict, json_path)
        request_body = safe_dict_lookup(
            data, 'requestBody', dict, json_path, None)
        return cls(
            operation_id=safe_dict_lookup(data, 'operationId', str, json_path),
            summary=safe_dict_lookup(data, 'summary', str, json_path, ''),
            method=method,
            parameters=[Parameter.from_dict(param, json_path + '.parameters.' + str(
                i), resolver) for i, param in enumerate(safe_dict_lookup(data, 'parameters', list, json_path, []))],
            request_body=RequestBody.from_dict(
                request_body, json_path + '.' + 'requestBody') if request_body is not None else None,
            responses=[Response.from_dict(k, v, json_path + '.' + k)
                       for k, v in safe_dict_lookup(data, 'responses', dict, json_path).items()],
            tags=set(safe_dict_lookup(data, 'tags', list, json_path))
        )

    def get_param_by_name(self, name: str) -> Parameter:
        for p in self.parameters:
            if p.name == name:
                return p
        raise Exception("Could not find parameter {} in operation {}".format(
            name, self.operation_id))


@dataclass
class Path:
    path: str
    operations: List[Operation]

    @classmethod
    def from_dict(cls: "Path", path_name: str, data: Any, json_path: str, resolver: Resolver) -> "Path":
        assert_type(data, dict, json_path)
        operations: List[Operation] = []
        for k, v in assert_type(data, dict, json_path).items():
            operations.append(Operation.from_dict(k, v, json_path + '.' + k, resolver))
        return cls(
            path=path_name,
            operations=operations
        )


@dataclass
class OpenApi:
    info: Info
    paths: List[Path]
    components: dict

    @classmethod
    def from_dict(cls: "OpenApi", data: Any) -> "OpenApi":
        assert_type(data, dict, '')
        resolver = Resolver(data)
        api = cls(
            info=Info.from_dict(safe_dict_lookup(
                data, 'info', dict, ''), 'info'),
            paths=[
                Path.from_dict(path_name, path_info, 'paths.' + path_name, resolver)
                for path_name, path_info in safe_dict_lookup(data, 'paths', dict, '').items()
                if not path_name.startswith('/packages')  # TODO: remove this
            ],
            components=safe_dict_lookup(data, 'components', dict, ''),
        )
        _resolve_links(api)
        return api


def _resolve_links(api: OpenApi):
    operations: Dict[str, Operation] = {}

    # first pass: collect all operations
    for path in api.paths:
        for operation in path.operations:
            if operation.operation_id in operations:
                raise Exception("Duplicate operation name {}".format(
                    operation.operation_id))
            operations[operation.operation_id] = operation

    # second pass: resolve links
    for path in api.paths:
        for operation in path.operations:
            for response in operation.responses:
                for link in response.links:
                    link.target_operation = operations[link.target_operation]
                    for param_name, param_query in link.parameters.items():
                        param = link.target_operation.get_param_by_name(
                            param_name)
                        if param.source_link:
                            raise Exception("Param already has a source link")
                        param.source_link = BackwardLink(
                            source_operation=operation,
                            source_response=response,
                            param_query=param_query,
                        )



from typing import List
from enum import Enum


class Level(Enum):
    INFO = 0
    WARNING = 1
    ERROR = 2

    def __or__(self, other: "Level") -> "Level":
        return Level(max(self.value, other.value))

    def color(self) -> str:
        if self.value == 0:
            return '\033[92m'
        if self.value == 1:
            return '\033[93m'
        if self.value == 2:
            return '\033[91m'
        return '\033[94m'


class AasTestResult:

    def __init__(self, message: str, path_fragment: str = '', level = Level.INFO):
        self.message = message
        self.path_fragment = path_fragment
        self.level = level
        self.sub_results: List[AasTestResult] = []

    def append(self, result: "AasTestResult"):
        self.sub_results.append(result)
        self.level = self.level | result.level

    def ok(self) -> bool:
        return self.level == Level.INFO

    def dump(self, indent=0, path=''):
        ENDC = '\033[0m'
        print("   " * indent + self.level.color() + self.message + ENDC)
        for sub_result in self.sub_results:
            sub_result.dump(indent + 1, path + "/" + self.path_fragment)

    def to_json(self):
        return {
            'm': self.message,
            'f': self.path_fragment,
            'l': self.level.value,
            's': [i.to_json() for i in self.sub_results]
        }

    @classmethod
    def from_json(self, data: dict) -> "AasTestResult":
        v = AasTestResult(
            data['m'], data['f'], Level(data['l'])
        )
        for i in data['s']:
            v.append(AasTestResult.from_json(i))
        return v